@@ -110,7 +110,7 @@ class EventNotifyHandler(BaseHTTPRequestHandler, EventNotifyHandlerBase):
 
     # pylint: disable=no-self-use, missing-docstring
     def log_event(self, seq, service_id, timestamp):
-        log.info(
+        log.debug(
             "Event %s received for %s service on thread %s at %s",
             seq,
             service_id,
@@ -145,7 +145,7 @@ class EventServerThread(threading.Thread):
         Handling of requests is delegated to an instance of the
         `EventNotifyHandler` class.
         """
-        log.info("Event listener running on %s", self.server.server_address)
+        log.debug("Event listener running on %s", self.server.server_address)
         # Listen for events until told to stop
         while not self.stop_flag.is_set():
             self.server.handle_request()
@@ -219,6 +219,7 @@ class EventListener(EventListenerBase):
         self._listener_thread.stop()
         # Send a dummy request in case the http server is currently listening
         try:
+            # pylint: disable=R1732
             urlopen("http://{}:{}/".format(address[0], address[1]))
         except URLError:
             # If the server is already shut down, we receive a socket error,
@@ -370,7 +371,7 @@ class Subscription(SubscriptionBase):
         self._auto_renew_thread_flag.set()
 
     # pylint: disable=no-self-use, too-many-arguments
-    def _request(self, method, url, headers, success):
+    def _request(self, method, url, headers, success, unconditional=None):
         """Sends an HTTP request.
 
         Args:
@@ -381,6 +382,9 @@ class Subscription(SubscriptionBase):
             success (function): A function to be called if the
                 request succeeds. The function will be called with a dict
                 of response headers as its only parameter.
+            unconditional (function): An optional function to be called after
+                the request is complete, regardless of its success. Takes
+                no parameters.
 
         """
         response = None
@@ -399,6 +403,8 @@ class Subscription(SubscriptionBase):
 
         if success:
             success(response.headers)
+        if unconditional:
+            unconditional()
 
     # pylint: disable=inconsistent-return-statements
     def _wrap(self, method, strict, *args, **kwargs):
